#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.logging.logging_util import get_logger

from octobot_services.interfaces.bots import abstract_bot_interface

from octobot_services.interfaces.bots.abstract_bot_interface import (
    AbstractBotInterface,
)

LOGGER = get_logger(__name__)
EOL = "\n"
NO_TRADER_MESSAGE = """OctoBot is either starting or there is no trader is activated in my config/config.json file.
See https://github.com/Drakkar-Software/OctoBot/wiki if you need help with my configuration."""
NO_CURRENCIES_MESSAGE = "No cryptocurrencies are in my activated profile.\n" \
                        "See https://github.com/Drakkar-Software/OctoBot/wiki/Configuration#cryptocurrencies " \
                        "if you need help with my cryptocurrencies configuration."""
UNAUTHORIZED_USER_MESSAGE = "Hello, I dont talk to strangers."

__all__ = [
    "AbstractBotInterface",
    "LOGGER",
    "EOL",
    "NO_TRADER_MESSAGE",
    "NO_CURRENCIES_MESSAGE"
]



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc
from octobot_commons import number_util

import octobot_commons.constants as common_constants
import octobot_commons.pretty_printer as pretty_printer
import octobot_commons.timestamp_util as timestamp_util

import octobot_trading.api as trading_api
import octobot_trading.constants as trading_constants

import octobot_services.interfaces as interfaces
import octobot_services.constants as constants
import octobot_commons.constants as commons_constants


class AbstractBotInterface(interfaces.AbstractInterface):
    __metaclass__ = abc.ABCMeta

    @staticmethod
    def enable(config, is_enabled, associated_config=None):
        if constants.CONFIG_INTERFACES not in config:
            config[constants.CONFIG_INTERFACES] = {}
        if associated_config not in config[constants.CONFIG_INTERFACES]:
            config[constants.CONFIG_INTERFACES][associated_config] = {}
        config[constants.CONFIG_INTERFACES][associated_config][common_constants.CONFIG_ENABLED_OPTION] = is_enabled

    @staticmethod
    def is_enabled(config, associated_config=None):
        return constants.CONFIG_INTERFACES in config \
               and associated_config in config[constants.CONFIG_INTERFACES] \
               and common_constants.CONFIG_ENABLED_OPTION in config[constants.CONFIG_INTERFACES][associated_config] \
               and config[constants.CONFIG_INTERFACES][associated_config][common_constants.CONFIG_ENABLED_OPTION]

    @staticmethod
    def _is_valid_user(user_name, associated_config=None):
        config_interface = interfaces.get_global_config()[constants.CONFIG_CATEGORY_SERVICES][associated_config]

        white_list = config_interface[constants.CONFIG_USERNAMES_WHITELIST] \
            if constants.CONFIG_USERNAMES_WHITELIST in config_interface else None

        is_valid = not white_list or user_name in white_list or f"@{user_name}" in white_list

        return is_valid, white_list

    @staticmethod
    def get_command_configuration(markdown=False):
        _, bold, code = pretty_printer.get_markers(markdown)
        message = f"{bold}My configuration:{bold}{interfaces.EOL}{interfaces.EOL}"

        message += f"{bold}Traders: {bold}{interfaces.EOL}"
        if interfaces.has_trader():
            has_real_trader, has_simulated_trader = interfaces.has_real_and_or_simulated_traders()
            if has_real_trader:
                message += f"{code}- Real trader{code}{interfaces.EOL}"
            if has_simulated_trader:
                message += f"{code}- Simulated trader{code}{interfaces.EOL}"
        else:
            message += f"{code}- No activated trader{code}{interfaces.EOL}"

        message += f"{interfaces.EOL}{bold}Exchanges:{bold}{interfaces.EOL}"
        for exchange_name in trading_api.get_exchange_names():
            message += f"{code}- {exchange_name.capitalize()}{code}{interfaces.EOL}"

        try:
            import octobot_evaluators.api as evaluators_api
            import octobot_evaluators.enums as evaluators_enums
            tentacle_setup_config = interfaces.get_bot_api().get_tentacles_setup_config()
            message += f"{interfaces.EOL}{bold}Evaluators:{bold}{interfaces.EOL}"
            evaluators = evaluators_api.get_evaluator_classes_from_type(
                evaluators_enums.EvaluatorMatrixTypes.TA.value, tentacle_setup_config)
            evaluators += evaluators_api.get_evaluator_classes_from_type(
                evaluators_enums.EvaluatorMatrixTypes.SOCIAL.value, tentacle_setup_config)
            evaluators += evaluators_api.get_evaluator_classes_from_type(
                evaluators_enums.EvaluatorMatrixTypes.REAL_TIME.value, tentacle_setup_config)
            for evaluator in evaluators:
                message += f"{code}- {evaluator.get_name()}{code}{interfaces.EOL}"

            message += f"{interfaces.EOL}{bold}Strategies:{bold}{interfaces.EOL}"
            for strategy in evaluators_api.get_evaluator_classes_from_type(
                    evaluators_enums.EvaluatorMatrixTypes.STRATEGIES.value, tentacle_setup_config):
                message += f"{code}- {strategy.get_name()}{code}{interfaces.EOL}"
        except ImportError:
            message += f"{interfaces.EOL}{bold}Impossible to retrieve evaluation configuration: requires OctoBot-Evaluators " \
                       f"package installed{bold}{interfaces.EOL}"
        try:
            trading_mode = interfaces.get_bot_api().get_trading_mode()
        except IndexError:
            # no activated trader
            trading_mode = None
        if trading_mode:
            message += f"{interfaces.EOL}{bold}Trading mode:{bold}{interfaces.EOL}"
            message += f"{code}- {trading_mode.get_name()}{code}"

        return message

    @staticmethod
    def get_command_market_status(markdown=False):
        _, bold, code = pretty_printer.get_markers(markdown)
        message = f"{bold}My cryptocurrencies evaluations are:{bold} {interfaces.EOL}{interfaces.EOL}"
        at_least_one_currency = False
        for currency_pair, currency_info in interfaces.get_currencies_with_status().items():
            at_least_one_currency = True
            message += f"{code}{currency_pair}:{code}{interfaces.EOL}"
            for _, evaluation in currency_info.items():
                message += f"{code}- {evaluation[2].capitalize()}: {evaluation[0]}{code}{interfaces.EOL}"
        if not at_least_one_currency:
            message += f"{code}{interfaces.NO_CURRENCIES_MESSAGE}{code}{interfaces.EOL}"
        risk = interfaces.get_risk()
        if risk:
            message += f"{interfaces.EOL}{code}My current risk is: {interfaces.get_risk()}{code}"
        return message

    @staticmethod
    def _print_trades(trades_history, trader_str, markdown=False):
        _, bold, code = pretty_printer.get_markers(markdown)
        trades_history_string = f"{bold}{trader_str}{bold}{code}Trades :{interfaces.EOL}{code}"
        if trades_history:
            for trade in trades_history:
                exchange_name = trading_api.get_trade_exchange_name(trade)
                trades_history_string += \
                    f"{pretty_printer.trade_pretty_printer(exchange_name, trade, markdown=markdown)}{interfaces.EOL}"
        else:
            trades_history_string += f"{code}No trade yet.{code}"
        return trades_history_string

    @staticmethod
    def get_command_trades_history(markdown=False):
        has_real_trader, has_simulated_trader = interfaces.has_real_and_or_simulated_traders()
        real_trades_history, simulated_trades_history = interfaces.get_trades_history()

        trades_history_string = ""
        if has_real_trader:
            trades_history_string += AbstractBotInterface._print_trades(real_trades_history,
                                                                        trading_constants.REAL_TRADER_STR,
                                                                        markdown)

        if has_simulated_trader:
            trades_history_string += f"{interfaces.EOL}" \
                                     f"{AbstractBotInterface._print_trades(simulated_trades_history, trading_constants.SIMULATOR_TRADER_STR, markdown)}"

        if not trades_history_string:
            trades_history_string = interfaces.NO_TRADER_MESSAGE

        return trades_history_string

    @staticmethod
    def _print_open_orders(open_orders, trader_str, markdown=False):
        _, bold, code = pretty_printer.get_markers(markdown)
        orders_string = f"{bold}{trader_str}{bold}{code}Open orders :{code}{interfaces.EOL}"
        if open_orders:
            for order in open_orders:
                exchange_name = trading_api.get_order_exchange_name(order).capitalize()
                orders_string += pretty_printer.open_order_pretty_printer(exchange_name,
                                                                          trading_api.order_to_dict(order),
                                                                          markdown=markdown) + interfaces.EOL
        else:
            orders_string += f"{code}No open order yet.{code}"
        return orders_string

    @staticmethod
    def get_command_open_orders(markdown=False):
        has_real_trader, has_simulated_trader = interfaces.has_real_and_or_simulated_traders()
        portfolio_real_open_orders, portfolio_simulated_open_orders = interfaces.get_all_open_orders()

        orders_string = ""
        if has_real_trader:
            orders_string += AbstractBotInterface._print_open_orders(portfolio_real_open_orders,
                                                                     trading_constants.REAL_TRADER_STR,
                                                                     markdown)

        if has_simulated_trader:
            message = AbstractBotInterface._print_open_orders(portfolio_simulated_open_orders,
                                                              trading_constants.SIMULATOR_TRADER_STR,
                                                              markdown)
            orders_string += f"{interfaces.EOL}{message}"

        if not orders_string:
            orders_string = interfaces.NO_TRADER_MESSAGE

        return orders_string

    @staticmethod
    def get_command_fees(markdown=False):
        _, bold, _ = pretty_printer.get_markers(markdown)
        real_trader_fees, simulated_trader_fees = interfaces.get_total_paid_fees()
        result_str = ""
        if real_trader_fees is not None:
            result_str = f"{bold}{trading_constants.REAL_TRADER_STR}{bold}{constants.PAID_FEES_STR}: " \
                         f"{pretty_printer.pretty_print_dict(real_trader_fees, markdown=markdown)}"
        if simulated_trader_fees is not None:
            result_str = f"{result_str}\n{bold}{trading_constants.SIMULATOR_TRADER_STR}{bold}" \
                         f"{constants.PAID_FEES_STR}: " \
                         f"{pretty_printer.pretty_print_dict(simulated_trader_fees, markdown=markdown)}"
        if not result_str:
            result_str = interfaces.NO_TRADER_MESSAGE
        return result_str

    @staticmethod
    def get_command_sell_all_currencies():
        try:
            interfaces.cancel_all_open_orders()
            nb_created_orders = len(interfaces.sell_all_currencies())
            if nb_created_orders:
                return f"Currencies sold in {nb_created_orders} order{'s' if nb_created_orders > 1 else ''}."
            else:
                return "Nothing to sell."
        except Exception as e:
            return f"An error occurred: {e.__class__.__name__}"

    @staticmethod
    def get_command_sell_all(currency):
        try:
            interfaces.cancel_all_open_orders(currency)
            nb_created_orders = len(interfaces.sell_all(currency))
            if nb_created_orders:
                return f"{currency} sold in {nb_created_orders} order{'s' if nb_created_orders > 1 else ''}."
            else:
                return f"Nothing to sell for {currency}."
        except Exception as e:
            return f"An error occurred: {e.__class__.__name__}"

    @staticmethod
    def _print_portfolio(
        current_val, ref_market, portfolio, currency_values, trader_str, markdown=False
    ):
        _, bold, code = pretty_printer.get_markers(markdown)
        portfolios_string = (
            f"{bold}{trader_str}{bold}Portfolio value : "
            f"{bold}{pretty_printer.get_min_string_from_number(current_val)} {ref_market}{bold}"
            f"{interfaces.EOL}"
        )
        portfolio_str = pretty_printer.global_portfolio_pretty_print(
            global_portfolio=portfolio, currency_values=currency_values, 
            ref_market_name=ref_market, markdown=markdown)

        if not portfolio_str:
            portfolio_str = "Nothing there."
        portfolios_string += f"{interfaces.EOL}{code}{portfolio_str}{code}"
        return portfolios_string

    @staticmethod
    def get_command_portfolio(markdown=False):
        has_real_trader, has_simulated_trader, \
        portfolio_real_current_value, portfolio_simulated_current_value = interfaces.get_portfolio_current_value()
        reference_market = interfaces.get_reference_market()
        real_global_portfolio, simulated_global_portfolio = interfaces.get_global_portfolio_currencies_amounts()
        currency_values = interfaces.get_global_portfolio_currencies_values()

        portfolios_string = ""
        if has_real_trader:
            portfolios_string += AbstractBotInterface._print_portfolio(
                portfolio_real_current_value,
                reference_market,
                real_global_portfolio,
                currency_values,
                trading_constants.REAL_TRADER_STR,
                markdown,
            )

        if has_simulated_trader:
            portfolio_str = AbstractBotInterface._print_portfolio(
                portfolio_simulated_current_value,
                reference_market,
                simulated_global_portfolio,
                currency_values,
                trading_constants.SIMULATOR_TRADER_STR,
                markdown,
            )
            portfolios_string += f"{interfaces.EOL}{portfolio_str}"

        if not portfolios_string:
            portfolios_string = interfaces.NO_TRADER_MESSAGE

        return portfolios_string

    @staticmethod
    def get_command_profitability(markdown=False):
        _, bold, code = pretty_printer.get_markers(markdown)
        has_real_trader, has_simulated_trader, \
        real_global_profitability, simulated_global_profitability, \
        real_percent_profitability, simulated_percent_profitability, \
        real_no_trade_profitability, simulated_no_trade_profitability, \
        market_average_profitability = interfaces.get_global_profitability()
        profitability_string = ""
        if has_real_trader:
            real_profitability_pretty = pretty_printer.portfolio_profitability_pretty_print(
                real_global_profitability, None, interfaces.get_reference_market())
            profitability_string = \
                f"{bold}{trading_constants.REAL_TRADER_STR}{bold}Global profitability : {code}{real_profitability_pretty}" \
                f"({pretty_printer.get_min_string_from_number(real_percent_profitability, 2)}%){code}, market: {code}" \
                f"{pretty_printer.get_min_string_from_number(market_average_profitability, 2)}%{code}, initial portfolio:" \
                f" {code}{pretty_printer.get_min_string_from_number(real_no_trade_profitability, 2)}%{code}{interfaces.EOL}"
        if has_simulated_trader:
            simulated_profitability_pretty = \
                pretty_printer.portfolio_profitability_pretty_print(
                    simulated_global_profitability, None, interfaces.get_reference_market())
            profitability_string += \
                f"{bold}{trading_constants.SIMULATOR_TRADER_STR}{bold}Global profitability : {code}" \
                f"{simulated_profitability_pretty}" \
                f"({pretty_printer.get_min_string_from_number(simulated_percent_profitability, 2)}%){code}, " \
                f"market: {code}{pretty_printer.get_min_string_from_number(market_average_profitability, 2)}%{code}, " \
                f"initial portfolio: {code}" \
                f"{pretty_printer.get_min_string_from_number(simulated_no_trade_profitability, 2)}%{code}"
        if not profitability_string:
            profitability_string = interfaces.NO_TRADER_MESSAGE

        return profitability_string

    @staticmethod
    def get_command_ping():
        return f"I'm alive since " \
               f"{timestamp_util.convert_timestamp_to_datetime(interfaces.get_bot_api().get_start_time(), '%Y-%m-%d %H:%M:%S')}."

    @staticmethod
    def get_command_version():
        return f"{interfaces.AbstractInterface.project_name} {interfaces.AbstractInterface.project_version}"

    @staticmethod
    def get_command_start(markdown=False):
        if markdown:
            return "Hello, I'm [OctoBot](https://github.com/Drakkar-Software/OctoBot), type /help to know my skills."
        else:
            return "Hello, I'm OctoBot, type /help to know my skills."

    @staticmethod
    def set_command_portfolios_refresh():
        return interfaces.trigger_portfolios_refresh()

    @staticmethod
    def set_command_risk(new_risk):
        return interfaces.set_risk(new_risk)

    @staticmethod
    def set_command_stop():
        interfaces.get_bot_api().stop_bot()

    def set_command_pause(self):
        interfaces.cancel_all_open_orders()
        interfaces.set_enable_trading(False)
        self.paused = True

    def set_command_resume(self):
        interfaces.set_enable_trading(True)
        self.paused = False

    @staticmethod
    def set_command_restart():
        interfaces.get_bot_api().restart_bot()

    @staticmethod
    def _split_messages_if_too_long(message, max_length, preferred_separator):
        if len(message) >= max_length:
            # split message using preferred_separator as separator
            messages_list = []
            first_part = message[:max_length]
            end_index = first_part.rfind(preferred_separator)
            if end_index != -1:
                messages_list.append(message[:end_index])
            else:
                messages_list.append(message[:max_length])
                end_index = len(first_part) - 1

            if end_index < len(message) - 1:
                remaining = message[end_index + 1:]
                return messages_list + AbstractBotInterface._split_messages_if_too_long(remaining, max_length,
                                                                                        preferred_separator)
            else:
                return messages_list
        else:
            return [message]



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_services.api import services
from octobot_services.api import interfaces
from octobot_services.api import service_feeds
from octobot_services.api import notification

from octobot_services.api.services import (
    get_available_services,
    get_service,
    create_service_factory,
    stop_services,
)
from octobot_services.api.interfaces import (
    initialize_global_project_data,
    create_interface_factory,
    is_enabled,
    is_enabled_in_backtesting,
    is_interface_relevant,
    disable_interfaces,
    send_user_command,
    start_interfaces,
    stop_interfaces,
)
from octobot_services.api.service_feeds import (
    create_service_feed_factory,
    get_service_feed,
    start_service_feed,
    stop_service_feed,
    clear_bot_id_feeds,
)
from octobot_services.api.notification import (
    create_notifier_factory,
    create_notification,
    is_enabled_in_config,
    get_enable_notifier,
    set_enable_notifier,
    is_notifier_relevant,
    send_notification,
    process_pending_notifications,
)


LOGGER_TAG = "ServicesApi"

__all__ = [
    "get_available_services",
    "get_service",
    "create_service_factory",
    "stop_services",
    "initialize_global_project_data",
    "create_interface_factory",
    "is_enabled",
    "is_enabled_in_backtesting",
    "is_interface_relevant",
    "disable_interfaces",
    "send_user_command",
    "start_interfaces",
    "stop_interfaces",
    "create_service_feed_factory",
    "get_service_feed",
    "start_service_feed",
    "stop_service_feed",
    "clear_bot_id_feeds",
    "create_notifier_factory",
    "create_notification",
    "is_enabled_in_config",
    "get_enable_notifier",
    "set_enable_notifier",
    "is_notifier_relevant",
    "send_notification",
    "process_pending_notifications",
]



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_services.service_feeds as service_feeds


class ServiceFeedManager:

    @staticmethod
    async def start_service_feed(service_feed: service_feeds.AbstractServiceFeed,
                                 backtesting_enabled: bool,
                                 edited_config: dict):
        if not service_feed.is_running and not service_feed.should_stop:
            if await service_feed.initialize(backtesting_enabled, edited_config):
                return await service_feed.start()
        return False

    @staticmethod
    async def stop_service_feed(service_feed: service_feeds.AbstractServiceFeed):
        logging.get_logger(ServiceFeedManager.__name__).debug(f"Stopping {service_feed.get_name()} ...")
        await service_feed.stop()
        logging.get_logger(ServiceFeedManager.__name__).debug(f"Stopped {service_feed.get_name()}")



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_services.notification import notification
from octobot_services.notification.notification import (
    Notification,
)
from octobot_services.notification import formated_notifications

from octobot_services.notification.formated_notifications import (
    OrderCreationNotification,
    OrderEndNotification,
)

__all__ = [
    "OrderCreationNotification",
    "OrderEndNotification",
    "Notification",
]



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.api as trading_api
import octobot_trading.errors as trading_errors

import octobot_commons.logging as logging

import octobot_services.interfaces as interfaces


def get_all_open_orders():
    simulated_open_orders = []
    real_open_orders = []

    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            if trading_api.is_trader_simulated(exchange_manager):
                simulated_open_orders += trading_api.get_open_orders(exchange_manager)
            else:
                real_open_orders += trading_api.get_open_orders(exchange_manager)

    return real_open_orders, simulated_open_orders


def cancel_orders(order_ids):
    removed_count = 0
    if order_ids:
        for order_id in order_ids:
            for exchange_manager in interfaces.get_exchange_managers():
                if trading_api.is_trader_existing_and_enabled(exchange_manager):
                    try:
                        removed_count += 1 if interfaces.run_in_bot_main_loop(
                            trading_api.cancel_order_with_id(exchange_manager, order_id, wait_for_cancelling=False)
                        ) else 0
                    except (trading_errors.OrderCancelError, trading_errors.UnexpectedExchangeSideOrderStateError) \
                            as err:
                        logging.get_logger("InterfaceOrderUtil").error(f"Skipping order cancel: {err}")
    return removed_count


def cancel_all_open_orders(currency=None):
    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            if currency is None:
                interfaces.run_in_bot_main_loop(
                    trading_api.cancel_all_open_orders(exchange_manager))
            else:
                interfaces.run_in_bot_main_loop(
                    trading_api.cancel_all_open_orders_with_currency(exchange_manager, currency))



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc

import octobot_commons.logging as logging

import octobot_services.services as services
import octobot_services.util as util


class AbstractServiceUser(util.InitializableWithPostAction):
    __metaclass__ = abc.ABCMeta

    # The service required to run this user
    REQUIRED_SERVICES = None

    def __init__(self, config):
        super().__init__()
        self.config = config
        self.paused = False

    async def _initialize_impl(self, backtesting_enabled, edited_config) -> bool:
        # init associated service if not already init
        service_list = services.ServiceFactory.get_available_services()
        if self.REQUIRED_SERVICES:
            for service in self.REQUIRED_SERVICES:
                if service in service_list:
                    if not await self._create_or_get_service_instance(service, backtesting_enabled, edited_config):
                        return False
                else:
                    self.get_logger().error(f"Required service {self.REQUIRED_SERVICES} is not an available service")
            return True
        elif self.REQUIRED_SERVICES is None:
            self.get_logger().error(f"Required service is not set, set it at False if no service is required")
        return False

    async def _create_or_get_service_instance(self, service, backtesting_enabled, edited_config):
        service_factory = services.ServiceFactory(self.config)
        if await service_factory.create_or_get_service(service, backtesting_enabled, edited_config):
            return True
        else:
            log_func = self.get_logger().debug
            # log error when the issue is not related to configuration
            if service.instance().has_required_configuration():
                log_func = self.get_logger().warning
            log_func(f"Impossible to start {self.get_name()}: required service {service.get_name()} is not available.")
            return False

    def has_required_services_configuration(self):
        return all(service.instance().has_required_configuration() for service in self.REQUIRED_SERVICES)

    @classmethod
    def get_name(cls):
        return cls.__name__

    @classmethod
    def get_logger(cls):
        return logging.get_logger(cls.get_name())



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_services.channel import abstract_service_feed
from octobot_services.channel import notifications

from octobot_services.channel.abstract_service_feed import (
    AbstractServiceFeedChannelConsumer,
    AbstractServiceFeedChannelProducer,
    AbstractServiceFeedChannel,
)
from octobot_services.channel.notifications import (
    NotificationChannelConsumer,
    NotificationChannelProducer,
    NotificationChannel,
)
from octobot_services.channel.user_commands import (
    UserCommandsChannelConsumer,
    UserCommandsChannelProducer,
    UserCommandsChannel,
)

__all__ = [
    "AbstractServiceFeedChannelConsumer",
    "AbstractServiceFeedChannelProducer",
    "AbstractServiceFeedChannel",
    "NotificationChannelConsumer",
    "NotificationChannelProducer",
    "NotificationChannel",
    "UserCommandsChannelConsumer",
    "UserCommandsChannelProducer",
    "UserCommandsChannel",
]




#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging

import octobot_trading.api as trading_api

import octobot_services.interfaces as interfaces


def has_trader():
    try:
        return trading_api.has_trader(_first_exchange_manager())
    except StopIteration:
        return False


def has_real_and_or_simulated_traders():
    has_real_trader = False
    has_simulated_trader = False
    exchange_managers = interfaces.get_exchange_managers()
    for exchange_manager in exchange_managers:
        if trading_api.is_trader_simulated(exchange_manager):
            has_simulated_trader = True
        else:
            has_real_trader = True
    return has_real_trader, has_simulated_trader


def sell_all_currencies():
    orders = []
    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            orders += interfaces.run_in_bot_main_loop(
                trading_api.sell_all_everything_for_reference_market(exchange_manager))
    return orders


def sell_all(currency):
    orders = []
    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            orders += interfaces.run_in_bot_main_loop(
                trading_api.sell_currency_for_reference_market(exchange_manager, currency))
    return orders


def set_enable_trading(enable):
    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.has_trader(exchange_manager):
            if trading_api.is_trader_enabled_in_config_from_exchange_manager(exchange_manager):
                trading_api.set_trading_enabled(exchange_manager, enable)


def _merge_trader_fees(current_fees, exchange_manager):
    current_fees_dict = current_fees if current_fees else {}
    for key, val in trading_api.get_total_paid_trading_fees(exchange_manager).items():
        if key in current_fees_dict:
            current_fees_dict[key] += val
        else:
            current_fees_dict[key] = val
    return current_fees_dict


def get_total_paid_fees(bot=None):
    real_trader_fees = None
    simulated_trader_fees = None

    for exchange_manager in interfaces.get_exchange_managers(bot):
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            if trading_api.is_trader_simulated(exchange_manager):
                simulated_trader_fees = _merge_trader_fees(simulated_trader_fees, exchange_manager)
            else:
                real_trader_fees = _merge_trader_fees(real_trader_fees, exchange_manager)
    return real_trader_fees, simulated_trader_fees


def get_trades_history(bot_api=None, symbol=None, independent_backtesting=None, since=None, as_dict=False):
    simulated_trades_history = []
    real_trades_history = []

    for exchange_manager in interfaces.get_exchange_managers(bot_api=bot_api,
                                                             independent_backtesting=independent_backtesting):
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            if trading_api.is_trader_simulated(exchange_manager):
                simulated_trades_history += trading_api.get_trade_history(exchange_manager, None, symbol, since, as_dict)
            else:
                real_trades_history += trading_api.get_trade_history(exchange_manager, None, symbol, since, as_dict)
    return real_trades_history, simulated_trades_history


def set_risk(risk):
    result_risk = None
    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.has_trader(exchange_manager):
            result_risk = trading_api.set_trader_risk(exchange_manager, risk)
    return result_risk


def get_risk():
    try:
        return trading_api.get_trader_risk(_first_exchange_manager()) \
            if trading_api.has_trader(_first_exchange_manager()) else None
    except StopIteration:
        return None


def get_currencies_with_status():
    evaluations_by_exchange_by_pair = {}
    for exchange_manager in interfaces.get_exchange_managers():
        trading_modes = trading_api.get_trading_modes(exchange_manager)
        for pair in trading_api.get_trading_pairs(exchange_manager):
            if pair not in evaluations_by_exchange_by_pair:
                evaluations_by_exchange_by_pair[pair] = {}
            status_explanation = "N/A"
            status = "N/A"
            for trading_mode in trading_modes:
                if trading_api.get_trading_mode_symbol(trading_mode) == pair:
                    status_explanation, status = trading_api.get_trading_mode_current_state(trading_mode)
                    try:
                        status = round(status, 3)
                    except TypeError:
                        pass
                    break
            evaluations_by_exchange_by_pair[pair][trading_api.get_exchange_manager_id(exchange_manager)] = \
                [status_explanation.replace("_", " "), status,
                 trading_api.get_exchange_name(exchange_manager).capitalize()]
    return evaluations_by_exchange_by_pair


def _get_tentacles_values(evaluations, tentacle_type_node, exchange):
    try:
        import octobot_evaluators.api as evaluators_api
    except ImportError:
        logging.get_logger("InterfaceUtil").error("_get_tentacles_values requires OctoBot-Evaluators package installed")
        return {}
    for tentacle_name, tentacle_name_node in evaluators_api.get_children_list(tentacle_type_node).items():
        evaluations[exchange][tentacle_name] = {}
        for cryptocurrency, cc_node in evaluators_api.get_children_list(tentacle_name_node).items():
            evaluations[exchange][tentacle_name][cryptocurrency] = {}
            if evaluators_api.has_children(cc_node):
                for symbol, symbol_node in evaluators_api.get_children_list(cc_node).items():
                    if evaluators_api.has_children(symbol_node):
                        evaluations[exchange][tentacle_name][symbol] = {}
                        for time_frame, time_frame_node in evaluators_api.get_children_list(symbol_node).items():
                            evaluations[exchange][tentacle_name][symbol][time_frame] = \
                                evaluators_api.get_value(time_frame_node)
                    else:
                        evaluations[exchange][tentacle_name][symbol] = evaluators_api.get_value(symbol_node)
            else:
                evaluations[exchange][tentacle_name][cryptocurrency] = evaluators_api.get_value(cc_node)


def get_matrix_list():
    try:
        import octobot_evaluators.api as evaluators_api
    except ImportError:
        logging.get_logger("InterfaceUtil").error("get_matrix_list requires OctoBot-Evaluators package installed")
        return {}
    evaluations = {}
    matrix = evaluators_api.get_matrix(interfaces.get_bot_api().get_matrix_id())
    for exchange, exchange_node in evaluators_api.get_node_children_by_names(matrix).items():
        evaluations[exchange] = {}
        for tentacle_type_node in evaluators_api.get_children_list(exchange_node).values():
            _get_tentacles_values(evaluations, tentacle_type_node, exchange)
    return evaluations


def _first_exchange_manager():
    return next(iter(interfaces.get_exchange_managers()))



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import threading

import octobot_commons.logging as logging
import octobot_commons.constants as commons_constants

import octobot_trading.api as trading_api

import octobot_services.interfaces as interfaces


def get_exchange_managers(bot_api=None, independent_backtesting=None, trading_exchanges_only=True):
    if bot_api is not None:
        return _filter_exchange_manager(trading_api.get_exchange_managers_from_exchange_ids(
            bot_api.get_exchange_manager_ids()), trading_exchanges_only)
    elif independent_backtesting is not None:
        try:
            import octobot.api as api
            return _filter_exchange_manager(
                trading_api.get_exchange_managers_from_exchange_ids(
                    api.get_independent_backtesting_exchange_manager_ids(independent_backtesting)),
                trading_exchanges_only)
        except ImportError:
            logging.get_logger("octobot_services/interfaces/util/util.py").error(
                "get_exchange_managers requires OctoBot package installed")
    else:
        return _filter_exchange_manager(interfaces.AbstractInterface.get_exchange_managers(), trading_exchanges_only)


def _filter_exchange_manager(exchange_managers, trading_exchanges_only):
    if trading_exchanges_only:
        return trading_api.get_trading_exchanges(exchange_managers)
    return exchange_managers


def run_in_bot_main_loop(coroutine, blocking=True, log_exceptions=True,
                         timeout=commons_constants.DEFAULT_FUTURE_TIMEOUT):
    if blocking:
        return interfaces.get_bot_api().run_in_main_asyncio_loop(coroutine, log_exceptions=log_exceptions,
                                                                 timeout=timeout)
    else:
        threading.Thread(
            target=interfaces.get_bot_api().run_in_main_asyncio_loop,
            name=f"run_in_bot_main_loop {coroutine.__name__}",
            args=(coroutine,),
            kwargs={"timeout": timeout}
         ).start()


def run_in_bot_async_executor(coroutine):
    return interfaces.get_bot_api().run_in_async_executor(coroutine)



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import abc

import octobot_commons.logging as logging


class ReturningStartable:
    __metaclass__ = abc.ABCMeta

    # Override this method with the actions to perform when starting this
    # Called both by async and threaded versions of this (in a threaded new async loop for threaded versions)
    @abc.abstractmethod
    async def _async_run(self) -> bool:
        raise NotImplementedError(f"_async_run is not implemented for {self.__class__.__name__}")

    # Override this method if this has to be run in a thread using this body
    # (with this extending threading.Thread)
    #
    # async def _inner_start(self) -> bool:
    #   threading.Thread.start(self)
    #   return True
    async def _inner_start(self) -> bool:
        return await self._async_run()

    # Always called to start this
    async def start(self) -> bool:
        try:
            return await self._inner_start()
        except Exception as e:
            class_name = self.__class__.__name__
            logger = logging.get_logger(class_name)
            logger.exception(e, True, f"{class_name} start error: {e}")
            return False

    # Called by threading.Thread.start(self) when a this is threaded
    def run(self) -> None:
        asyncio.run(self._async_run())



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

import octobot_commons.singleton as singleton

import octobot_services.service_feeds as service_feeds


class ServiceFeeds(singleton.Singleton):
    def __init__(self):
        self.service_feeds = {}

    def get_service_feed(self, bot_id: str, feed_name: str) -> typing.Optional[service_feeds.AbstractServiceFeed]:
        try:
            return self.service_feeds[bot_id][feed_name]
        except KeyError:
            return None

    def add_service_feed(self, bot_id: str, feed_name: str, feed: service_feeds.AbstractServiceFeed) -> None:
        if bot_id not in self.service_feeds:
            self.service_feeds[bot_id] = {}
        self.service_feeds[bot_id][feed_name] = feed

    def clear_bot_id_feeds(self, bot_id: str) -> None:
        self.service_feeds.pop(bot_id, None)



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc

import octobot_commons.enums as common_enums
import octobot_commons.pretty_printer as pretty_printer

import octobot_trading.constants as constants

import octobot_services.notification as notifications
import octobot_services.enums as enums


class _OrderNotification(notifications.Notification):
    def __init__(self, text, evaluator_notification: notifications.Notification):
        super().__init__("", text, "", enums.NotificationSound.NO_SOUND,
                         common_enums.MarkdownFormat.IGNORE,
                         enums.NotificationLevel.INFO,
                         enums.NotificationCategory.TRADES,
                         evaluator_notification)
        self._build_text()

    @abc.abstractmethod
    def _build_text(self):
        raise NotImplementedError("_build_text is not implemented")


class OrderCreationNotification(_OrderNotification):
    def __init__(self, evaluator_notification: notifications.Notification, dict_order: dict, exchange_name: str):
        self.dict_order = dict_order
        self.exchange_name = exchange_name.capitalize()
        super().__init__("Order creation", evaluator_notification)

    def _build_text(self):
        self.text = ""
        self.markdown_text = ""
        self.text += f"- {pretty_printer.open_order_pretty_printer(self.exchange_name, self.dict_order, markdown=False)}"
        self.markdown_text += \
            f"- {pretty_printer.open_order_pretty_printer(self.exchange_name, self.dict_order, markdown=True)}"


class OrderEndNotification(_OrderNotification):
    def __init__(self, order_previous_notification: notifications.Notification,
                 dict_order_filled: dict, exchange_name: str,
                 dict_orders_canceled: list, trade_profitability: float, portfolio_profitability: float,
                 portfolio_diff: float, add_profitability: bool, is_simulated: bool):
        self.dict_order_filled = dict_order_filled
        self.exchange_name = exchange_name.capitalize()
        self.dict_orders_canceled = dict_orders_canceled
        self.trade_profitability = trade_profitability
        self.portfolio_profitability = portfolio_profitability
        self.portfolio_diff = portfolio_diff
        self.add_profitability = add_profitability
        self.is_simulated = is_simulated
        super().__init__("Order update", order_previous_notification)

    def _build_text(self):
        self.text = ""
        self.markdown_text = ""
        trader_type = constants.SIMULATOR_TRADER_STR if self.is_simulated else constants.REAL_TRADER_STR
        if self.dict_order_filled is not None:
            self.text += f"{trader_type}Order(s) filled : " \
                         f"\n- {pretty_printer.open_order_pretty_printer(self.exchange_name, self.dict_order_filled)}"
            md_text = pretty_printer.open_order_pretty_printer(self.exchange_name, self.dict_order_filled, markdown=True)
            self.markdown_text += f"*{trader_type}*Order(s) filled : \n-{md_text}"

        if self.dict_orders_canceled is not None and self.dict_orders_canceled:
            self.text += f"{trader_type}Order(s) canceled :"
            self.markdown_text += f"*{trader_type}*Order(s) canceled :"
            for dict_order in self.dict_orders_canceled:
                self.text += f"\n- {pretty_printer.open_order_pretty_printer(self.exchange_name, dict_order)}"
                self.markdown_text += \
                    f"\n- {pretty_printer.open_order_pretty_printer(self.exchange_name, dict_order, markdown=True)}"

        if self.trade_profitability is not None and self.add_profitability:
            self.text += f"\nTrade profitability : {'+' if self.trade_profitability >= 0 else ''}" \
                         f"{round(self.trade_profitability * 100, 4)}%"
            self.markdown_text += f"\nTrade profitability : *{'+' if self.trade_profitability >= 0 else ''}" \
                                  f"{round(self.trade_profitability * 100, 4)}%*"

        if self.portfolio_profitability is not None and self.add_profitability:
            self.text += f"\nPortfolio profitability : {round(self.portfolio_profitability, 4)}% " \
                         f"{'+' if self.portfolio_diff >= 0 else ''}{round(self.portfolio_diff, 4)}%"
            self.markdown_text += f"\nPortfolio profitability : `{round(self.portfolio_profitability, 4)}% " \
                                  f"{'+' if self.portfolio_diff >= 0 else ''}{round(self.portfolio_diff, 4)}%`"



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc

import async_channel.channels as channels

import octobot_commons.asyncio_tools as asyncio_tools

import octobot_services.abstract_service_user as abstract_service_user
import octobot_services.channel as service_channels
import octobot_services.util as util


class AbstractServiceFeed(abstract_service_user.AbstractServiceUser,
                          util.ReturningStartable,
                          service_channels.AbstractServiceFeedChannelProducer):
    __metaclass__ = abc.ABCMeta

    # Override FEED_CHANNEL with a dedicated channel
    FEED_CHANNEL = None

    # Set simulator class when available in order to use it in backtesting for this feed
    SIMULATOR_CLASS = None

    _SLEEPING_TIME_BEFORE_RECONNECT_ATTEMPT_SEC = 10
    DELAY_BETWEEN_STREAMS_QUERIES = 5
    REQUIRED_SERVICE_ERROR_MESSAGE = "Required services are not ready, service feed can't start"

    def __init__(self, config, main_async_loop, bot_id):
        abstract_service_user.AbstractServiceUser.__init__(self, config)
        service_channels.AbstractServiceFeedChannelProducer.__init__(self, channels.set_chan(self.FEED_CHANNEL(), None))
        self.feed_config = {}
        self.main_async_loop = main_async_loop
        self.bot_id = bot_id
        self.services = None
        self.should_stop = False

    # Override update_feed_config if any need in the extending feed
    def update_feed_config(self, config):
        pass

    # Override this method if the service feed implementation is using a dispatcher handled in the service layer
    # (ie: TelegramServiceFeed)
    @staticmethod
    def _get_service_layer_service_feed() -> object:
        return None

    # Override this method to specify the feed reception process
    @abc.abstractmethod
    async def _start_service_feed(self):
        raise NotImplementedError("start_dispatcher not implemented")

    @abc.abstractmethod
    def _something_to_watch(self):
        raise NotImplementedError("_something_to_watch not implemented")

    @abc.abstractmethod
    def _initialize(self):
        raise NotImplementedError("_initialize not implemented")

    async def _init_channel(self):
        channel = channels.get_chan(self.FEED_CHANNEL.get_name())
        await channel.register_producer(self)

    # Call _notify_consumers to send data to consumers
    def _notify_consumers(self, data):
        try:
            # send notification only if is a notification channel is running
            channels.get_chan(self.FEED_CHANNEL.get_name())
            asyncio_tools.run_coroutine_in_asyncio_loop(self.feed_send_coroutine(data), self.main_async_loop)
        except KeyError:
            self.logger.error("Can't send notification data: no initialized channel found")

    # Call _async_notify_consumers to send data to consumers (same as _notify_consumers but directly from async context)
    async def _async_notify_consumers(self, data):
        try:
            # send notification only if is a notification channel is running
            channels.get_chan(self.FEED_CHANNEL.get_name())
            await self.feed_send_coroutine(data)
        except KeyError:
            self.logger.error("Can't send notification data: no initialized channel found")

    async def feed_send_coroutine(self, data):
        await self.send(
            {
                "data": data
            }
        )

    async def _run(self, should_init=True):
        self.is_running = True
        service_level_service_feed_if_any = self._get_service_layer_service_feed()
        if self._something_to_watch():
            if should_init:
                self._initialize()
                await self._init_channel()
            if self.services is not None:
                for service in self.services:
                    if service_level_service_feed_if_any is not None \
                            and not service.is_running():
                        await service.start_service_feed()
            if not await self._start_service_feed():
                self.logger.warning("Nothing can be monitored even though there is something to watch"
                                    ", feed is going closing.")
        else:
            self.logger.info("Nothing to monitor, feed is closing.")
            self.is_running = False
        return True

    async def _async_run(self) -> bool:
        self.logger.info("Initializing feed reception ...")
        self.services = [service.instance() for service in self.REQUIRED_SERVICES]
        return await self._run()

    async def resume(self) -> bool:
        self.should_stop = False
        self.logger.info("Resuming feed reception ...")
        return await self._run(should_init=False)

    async def stop(self):
        if self.is_running:
            self.should_stop = True
            self.is_running = False



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.api as trading_api
import octobot_trading.enums as trading_enums

import octobot_services.interfaces as interfaces


def get_all_positions():
    simulated_positions = []
    real_positions = []

    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            if trading_api.is_trader_simulated(exchange_manager):
                simulated_positions += trading_api.get_positions(exchange_manager)
            else:
                real_positions += trading_api.get_positions(exchange_manager)

    return real_positions, simulated_positions


def close_positions(positions_descs):
    removed_count = 0
    if positions_descs:
        for positions_desc in positions_descs:
            for exchange_manager in interfaces.get_exchange_managers():
                if trading_api.is_trader_existing_and_enabled(exchange_manager):
                    removed_count += 1 if interfaces.run_in_bot_main_loop(
                        trading_api.close_position(
                            exchange_manager,
                            positions_desc["symbol"],
                            trading_enums.PositionSide(positions_desc["side"]),
                        )
                    ) else 0
    return removed_count



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_services.interfaces.util import bot
from octobot_services.interfaces.util import util
from octobot_services.interfaces.util import order
from octobot_services.interfaces.util import position
from octobot_services.interfaces.util import trader
from octobot_services.interfaces.util import portfolio
from octobot_services.interfaces.util import profitability

from octobot_services.interfaces.util.bot import (
    get_bot_api,
    get_exchange_manager_ids,
    get_global_config,
    get_startup_config,
    get_edited_config,
    get_startup_tentacles_config,
    get_edited_tentacles_config,
    set_edited_tentacles_config,
)
from octobot_services.interfaces.util.util import (
    get_exchange_managers,
    run_in_bot_main_loop,
    run_in_bot_async_executor,
)
from octobot_services.interfaces.util.order import (
    get_all_open_orders,
    cancel_orders,
    cancel_all_open_orders,
)
from octobot_services.interfaces.util.position import (
    get_all_positions,
    close_positions,
)
from octobot_services.interfaces.util.trader import (
    has_trader,
    has_real_and_or_simulated_traders,
    sell_all_currencies,
    sell_all,
    set_enable_trading,
    get_total_paid_fees,
    get_trades_history,
    set_risk,
    get_risk,
    get_currencies_with_status,
    get_matrix_list,
)
from octobot_services.interfaces.util.portfolio import (
    get_portfolio_holdings,
    get_portfolio_current_value,
    get_global_portfolio_currencies_amounts,
    get_global_portfolio_currencies_values,
    trigger_portfolios_refresh,
)
from octobot_services.interfaces.util.profitability import (
    get_global_profitability,
    get_reference_market,
)

__all__ = [
    "get_bot_api",
    "get_exchange_manager_ids",
    "get_global_config",
    "get_startup_config",
    "get_edited_config",
    "get_startup_tentacles_config",
    "get_edited_tentacles_config",
    "set_edited_tentacles_config",
    "get_exchange_managers",
    "run_in_bot_main_loop",
    "run_in_bot_async_executor",
    "get_all_open_orders",
    "cancel_orders",
    "cancel_all_open_orders",
    "has_trader",
    "has_real_and_or_simulated_traders",
    "sell_all_currencies",
    "sell_all",
    "set_enable_trading",
    "get_total_paid_fees",
    "get_trades_history",
    "set_risk",
    "get_risk",
    "get_currencies_with_status",
    "get_matrix_list",
    "get_portfolio_holdings",
    "get_portfolio_current_value",
    "get_global_portfolio_currencies_amounts",
    "get_global_portfolio_currencies_values",
    "trigger_portfolios_refresh",
    "get_global_profitability",
    "get_reference_market",
    "get_all_positions",
    "close_positions",
]



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_services.managers as managers
import octobot_services.services as services
import octobot_services.interfaces as interfaces
import octobot_services.errors as errors
import octobot_commons.asyncio_tools as asyncio_tools


_SERVICE_ASYNC_LOCKS = {}


def _service_async_lock(service_class):
    try:
        return _SERVICE_ASYNC_LOCKS[service_class.__name__]
    except KeyError:
        _SERVICE_ASYNC_LOCKS[service_class.__name__] = asyncio_tools.RLock()
        return _SERVICE_ASYNC_LOCKS[service_class.__name__]


def get_available_services() -> list:
    return services.ServiceFactory.get_available_services()


async def get_service(service_class, is_backtesting):
    # prevent concurrent access when creating a service
    async with _service_async_lock(service_class):
        if await create_service_factory(
                interfaces.get_startup_config(dict_only=True)
        ).create_or_get_service(service_class, is_backtesting, interfaces.get_edited_config(dict_only=True)):
            service = service_class.instance()
            if is_backtesting and not service.BACKTESTING_ENABLED:
                raise errors.UnavailableInBacktestingError(
                    f"{service_class.__name__} service is not available in backtesting"
                )
            return service
    raise errors.CreationError(f"{service_class.__name__} service has not been properly created")


def create_service_factory(config) -> services.ServiceFactory:
    return services.ServiceFactory(config)


async def stop_services() -> None:
    await managers.stop_services()



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.logging as logging

import octobot_services.constants as constants
import octobot_services.services as services


class ServiceFactory:
    def __init__(self, config):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.config = config

    @staticmethod
    def get_available_services() -> list:
        return [service_class for service_class in services.AbstractService.__subclasses__()]

    async def create_or_get_service(self, service_class, backtesting_enabled, edited_config) -> bool:
        """
        create_or_get_service will create a service instance if it doesn't exist, check the existing one otherwise
        :param service_class: the class of the service to create
        :return: True if the created service is working properly, False otherwise
        """
        service_instance = service_class.instance()
        if service_instance.get_has_been_created():
            return service_instance.is_healthy()
        else:
            return await self._create_service(service_instance, backtesting_enabled, edited_config)

    async def _create_service(self, service, backtesting_enabled, edited_config) -> bool:
        service.is_backtesting_enabled = backtesting_enabled
        service.set_has_been_created(True)
        service.logger = logging.get_logger(service.get_name())
        service.config = self.config
        service.edited_config = edited_config
        if service.has_required_configuration():
            return await self._perform_checkup(service)
        else:
            if service.get_should_warn():
                self.logger.info(f"{service.get_name()} can't be initialized: configuration "
                                 f"is missing, wrong or incomplete. This is normal if you did not configure it yet.")
        return False

    async def _perform_checkup(self, service) -> bool:
        try:
            await service.prepare()
            if constants.CONFIG_CATEGORY_SERVICES not in self.config:
                self.config[constants.CONFIG_CATEGORY_SERVICES] = {}
            if service.get_type() not in self.config[constants.CONFIG_CATEGORY_SERVICES]:
                self.config[constants.CONFIG_CATEGORY_SERVICES][service.get_type()] = {}
            self.config[constants.CONFIG_CATEGORY_SERVICES][service.get_type()][constants.CONFIG_SERVICE_INSTANCE] = \
                service
            if await service.say_hello():
                return service.is_healthy()
            else:
                self.logger.warning(f"{service.get_name()} initial checkup failed.")
        except Exception as e:
            self.logger.exception(e, True, f"{service.get_name()} preparation produced the following error: {e}")
        return False

    @staticmethod
    def has_already_been_created(service_class):
        return service_class.instance().get_has_been_created()



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import enum

import async_channel.channels as channels

import octobot_commons.channels_name as channels_name
import octobot_commons.logging as logging
import octobot_commons.enums as enums

import octobot_services.api as api
import octobot_services.managers as managers

OCTOBOT_CHANNEL_SERVICE_CONSUMER_LOGGER_TAG = "OctoBotChannelServiceConsumer"


class OctoBotChannelServiceActions(enum.Enum):
    """
    OctoBot Channel consumer supported actions
    """

    INTERFACE = "interface"
    NOTIFICATION = "notification"
    SERVICE_FEED = "service_feed"
    START_SERVICE_FEED = "start_service_feed"
    EXCHANGE_REGISTRATION = "exchange_registration"


class OctoBotChannelServiceDataKeys(enum.Enum):
    """
    OctoBot Channel consumer supported data keys
    """

    EXCHANGE_ID = "exchange_id"
    BOT_ID = "bot_id"
    EDITED_CONFIG = "edited_config"
    BACKTESTING_ENABLED = "backtesting_enabled"
    INSTANCE = "instance"
    SUCCESSFUL_OPERATION = "successful_operation"
    CLASS = "class"
    FACTORY = "factory"
    EXECUTORS = "executors"


async def octobot_channel_callback(bot_id, subject, action, data) -> None:
    """
    OctoBot channel consumer callback
    :param bot_id: the callback bot id
    :param subject: the callback subject
    :param action: the callback action
    :param data: the callback data
    """
    if subject == enums.OctoBotChannelSubjects.CREATION.value:
        await _handle_creation(bot_id, action, data)
    elif subject == enums.OctoBotChannelSubjects.UPDATE.value:
        if action == OctoBotChannelServiceActions.EXCHANGE_REGISTRATION.value:
            await _handle_exchange_notification(data)
        elif action == OctoBotChannelServiceActions.START_SERVICE_FEED.value:
            await _handle_service_feed_start_notification(bot_id, action, data)


async def _handle_creation(bot_id, action, data):
    created_instance = None
    edited_config = data[OctoBotChannelServiceDataKeys.EDITED_CONFIG.value]
    backtesting_enabled = data[OctoBotChannelServiceDataKeys.BACKTESTING_ENABLED.value]
    to_create_class = data[OctoBotChannelServiceDataKeys.CLASS.value]
    factory = data[OctoBotChannelServiceDataKeys.FACTORY.value]
    if action == OctoBotChannelServiceActions.INTERFACE.value:
        created_instance = await _create_and_start_interface(factory, to_create_class,
                                                             edited_config, backtesting_enabled)
    if action == OctoBotChannelServiceActions.NOTIFICATION.value:
        executors = data[OctoBotChannelServiceDataKeys.EXECUTORS.value]
        created_instance = await _create_notifier(factory, to_create_class, edited_config,
                                                  backtesting_enabled, executors)
    if action == OctoBotChannelServiceActions.SERVICE_FEED.value:
        created_instance = await _create_service_feed(factory, to_create_class)
    await channels.get_chan_at_id(channels_name.OctoBotChannelsName.OCTOBOT_CHANNEL.value,
                                  bot_id).get_internal_producer() \
        .send(bot_id=bot_id,
              subject=enums.OctoBotChannelSubjects.NOTIFICATION.value,
              action=action,
              data={OctoBotChannelServiceDataKeys.INSTANCE.value: created_instance})


async def _create_and_start_interface(interface_factory, to_create_class, edited_config, backtesting_enabled):
    interface_instance = await interface_factory.create_interface(to_create_class)
    await interface_instance.initialize(backtesting_enabled, edited_config)
    return interface_instance if await managers.start_interface(interface_instance) else None


async def _create_notifier(factory, to_create_class, edited_config, backtesting_enabled, executors):
    notifier_instance = await factory.create_notifier(to_create_class)
    notifier_instance.executors = executors
    await notifier_instance.initialize(backtesting_enabled, edited_config)
    return notifier_instance


async def _create_service_feed(factory, to_create_class):
    return factory.create_service_feed(to_create_class)


async def _handle_exchange_notification(data):
    notifier_or_interface = data[OctoBotChannelServiceDataKeys.INSTANCE.value]
    exchange_id = data[OctoBotChannelServiceDataKeys.EXCHANGE_ID.value]
    await notifier_or_interface.register_new_exchange(exchange_id)


async def _handle_service_feed_start_notification(bot_id, action, data):
    service_feed = data[OctoBotChannelServiceDataKeys.INSTANCE.value]
    edited_config = data[OctoBotChannelServiceDataKeys.EDITED_CONFIG.value]
    await channels.get_chan_at_id(channels_name.OctoBotChannelsName.OCTOBOT_CHANNEL.value,
                                  bot_id).get_internal_producer() \
        .send(bot_id=bot_id,
              subject=enums.OctoBotChannelSubjects.NOTIFICATION.value,
              action=action,
              data={OctoBotChannelServiceDataKeys.SUCCESSFUL_OPERATION.value: await _start_service_feed(service_feed,
                                                                                                        edited_config)})


async def _start_service_feed(service_feed, edited_config):
    if not await api.start_service_feed(service_feed, False, edited_config):
        logger = logging.get_logger(OCTOBOT_CHANNEL_SERVICE_CONSUMER_LOGGER_TAG)
        # log error when the issue is not related to configuration
        if service_feed.has_required_services_configuration():
            logger.error(f"Failed to start {service_feed.get_name()}. Evaluators requiring this service feed "
                         f"might not work properly.")
        else:
            logger.debug(f"Impossible to start {service_feed.get_name()}: missing service(s) configuration.")
        return False
    return True



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc

import octobot_commons.channels_name as channels_names
import async_channel.util as channel_creator
import async_channel.channels as channels
import octobot_services.channel as service_channels
import octobot_services.abstract_service_user as abstract_service_user
import octobot_services.util as util


class AbstractInterface(abstract_service_user.AbstractServiceUser, util.ReturningStartable, util.ExchangeWatcher):
    __metaclass__ = abc.ABCMeta
    # The service required to run this interface
    REQUIRED_SERVICES = None

    # References that will be shared by interfaces
    bot_api = None
    project_name = None
    project_version = None
    enabled = True

    def __init__(self, config):
        abstract_service_user.AbstractServiceUser.__init__(self, config)
        util.ExchangeWatcher.__init__(self)

    async def _initialize_impl(self, backtesting_enabled, edited_config) -> bool:
        if await abstract_service_user.AbstractServiceUser._initialize_impl(self, backtesting_enabled, edited_config):
            await self._create_user_commands_channel_if_not_existing()
            return True
        return False

    @staticmethod
    def initialize_global_project_data(bot_api, project_name, project_version):
        AbstractInterface.bot_api = bot_api
        AbstractInterface.project_name = project_name
        AbstractInterface.project_version = project_version

    @staticmethod
    def get_exchange_managers():
        try:
            import octobot_trading.api as api
            return api.get_exchange_managers_from_exchange_ids(AbstractInterface.bot_api.get_exchange_manager_ids())
        except ImportError:
            AbstractInterface.get_logger().error("AbstractInterface requires OctoBot-Trading package installed")

    @staticmethod
    def is_bot_ready():
        return AbstractInterface.bot_api.is_initialized()

    @abc.abstractmethod
    async def stop(self):
        raise NotImplementedError(f"stop is not implemented for {self.get_name()}")

    @staticmethod
    async def _create_user_commands_channel_if_not_existing() -> None:
        try:
            channels.get_chan(channels_names.OctoBotUserChannelsName.USER_COMMANDS_CHANNEL.value)
        except KeyError:
            await channel_creator.create_channel_instance(service_channels.UserCommandsChannel, channels.set_chan)



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging


async def start_interfaces(interfaces: list):
    started_interfaces = []
    for interface in interfaces:
        if await interface.start():
            started_interfaces.append(interface)
    return started_interfaces


async def start_interface(interface):
    return await interface.start()


async def stop_interfaces(interfaces: list):
    for interface in interfaces:
        logging.get_logger(__name__).debug(f"Stopping {interface.get_name()} ...")
        await interface.stop()
        logging.get_logger(__name__).debug(f"Stopped {interface.get_name()}")



#  Drakkar-Software OctoBot-Interfaces
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest

from octobot_services.api.interfaces import initialize_global_project_data, create_interface_factory, \
    start_interfaces, stop_interfaces
from octobot_services.interfaces.abstract_interface import AbstractInterface


def test_initialize_global_project_data():
    bot_api = "bot"
    initialize_global_project_data(bot_api, "1", "2")
    assert AbstractInterface.bot_api is bot_api
    assert AbstractInterface.project_name == "1"
    assert AbstractInterface.project_version == "2"


def test_create_interface_factory():
    create_interface_factory({})


@pytest.mark.asyncio
async def test_start_interfaces():
    await start_interfaces([])


@pytest.mark.asyncio
async def test_stop_interfaces():
    await stop_interfaces([])



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.channels as channel

import octobot_commons.enums as common_enums

import octobot_services.channel as channels
import octobot_services.notification as notifications
import octobot_services.notifier as notifiers
import octobot_services.enums as enums

MAX_PENDING_NOTIFICATION = 10
pending_notifications = []


def create_notifier_factory(config) -> notifiers.NotifierFactory:
    return notifiers.NotifierFactory(config)


def create_notification(text: str, title="", markdown_text="", sound=enums.NotificationSound.NO_SOUND,
                        markdown_format: common_enums.MarkdownFormat = common_enums.MarkdownFormat.IGNORE,
                        level: enums.NotificationLevel = enums.NotificationLevel.INFO,
                        category: enums.NotificationCategory = enums.NotificationCategory.GLOBAL_INFO,
                        linked_notification=None) -> notifications.Notification:
    return notifications.Notification(text, title, markdown_text, sound, markdown_format, level, category,
                                      linked_notification)


async def send_notification(notification: notifications.Notification) -> None:
    try:
        # send notification only if is a notification channel is running
        channel.get_chan(channels.NotificationChannel.get_name())
        await channels.NotificationChannelProducer.instance().send(
            {
                "notification": notification
            }
        )
    except KeyError:
        if len(pending_notifications) < MAX_PENDING_NOTIFICATION:
            pending_notifications.append(notification)


async def process_pending_notifications():
    for notification in pending_notifications:
        await channels.NotificationChannelProducer.instance().send(
            {
                "notification": notification
            }
        )
    pending_notifications.clear()


def is_enabled_in_config(notifier_class, config) -> bool:
    return notifier_class.is_enabled(config)


def get_enable_notifier(notifier) -> bool:
    return notifier.enabled


def set_enable_notifier(notifier, enabled) -> None:
    notifier.enabled = enabled


def is_notifier_relevant(config, notifier_class, backtesting_enabled):
    return is_enabled_in_config(notifier_class, config) and \
           all(service.get_is_enabled(config)
               for service in notifier_class.REQUIRED_SERVICES) and \
           not backtesting_enabled



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class ExchangeWatcher:
    """
    ExchangeWatcher is used as a superclass for elements that require to interact with exchanges.
    register_new_exchange_impl(self, exchange_id) will be called whenever a new exchange is ready.
    Registered exchange ids are stored in self.registered_exchanges_ids
    """
    def __init__(self):
        self.registered_exchanges_ids = set()

    async def register_new_exchange(self, exchange_id):
        try:
            await self.register_new_exchange_impl(exchange_id)
        finally:
            self.registered_exchanges_ids.add(exchange_id)

    async def register_new_exchange_impl(self, exchange_id):
        pass



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.tentacles_management as tentacles_management

import octobot_commons.logging as logging

import octobot_services.service_feeds as service_feeds


class ServiceFeedFactory:
    def __init__(self, config, main_async_loop, bot_id):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.config = config
        self.main_async_loop = main_async_loop
        self.bot_id = bot_id

    @staticmethod
    def get_available_service_feeds(in_backtesting: bool) -> list:
        feeds = tentacles_management.get_all_classes_from_parent(service_feeds.AbstractServiceFeed)
        if in_backtesting:
            feeds = [feed.SIMULATOR_CLASS
                     for feed in feeds
                     if feed.SIMULATOR_CLASS is not None]
        return feeds

    def create_service_feed(self, service_feed_class) -> service_feeds.AbstractServiceFeed:
        feed = service_feed_class(self.config, self.main_async_loop, self.bot_id)
        service_feeds.ServiceFeeds.instance().add_service_feed(self.bot_id, service_feed_class.get_name(), feed)
        return feed



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc
import asyncio

import async_channel.util as channel_creator
import async_channel.channels as channels

import octobot_trading.api as trading_api
import octobot_trading.enums as trading_enums

import octobot_services.channel as service_channels
import octobot_services.constants as constants
import octobot_services.notification as notifications
import octobot_services.abstract_service_user as abstract_service_user
import octobot_services.util as util


class AbstractNotifier(abstract_service_user.AbstractServiceUser, util.ExchangeWatcher):
    __metaclass__ = abc.ABCMeta
    # Override this key with the identifier of the notifier (used to know if enabled)
    NOTIFICATION_TYPE_KEY = None
    # The service required to run this notifier
    REQUIRED_SERVICES = None
    USE_MAIN_LOOP = False

    def __init__(self, config):
        abstract_service_user.AbstractServiceUser.__init__(self, config)
        util.ExchangeWatcher.__init__(self)
        self.executors = None
        self.logger = self.get_logger()
        self.enabled = self.is_enabled(config)
        self.services = None
        self.previous_notifications_by_identifier = {}
        self.loop = asyncio.get_event_loop()

    async def register_new_exchange_impl(self, exchange_id):
        # if self.is_initialized is False, this notifier has not been initialized and should not be used
        if self.is_initialized and exchange_id not in self.registered_exchanges_ids:
            await self._subscribe_to_order_channel(exchange_id)

    # Override this method to consume a notification when received
    @abc.abstractmethod
    async def _handle_notification(self, notification: notifications.Notification):
        raise NotImplementedError(f"_handle_notification is not implemented")

    async def _send_notification(self, notification: notifications.Notification):
        self.logger.debug(f"Publishing notification: {notification}")
        await self._handle_notification(notification)

    def _send_notification_from_executor(self, notification: notifications.Notification):
        asyncio.run(self._send_notification(notification))

    async def _notification_callback(self, notification: notifications.Notification = None):
        try:
            if self._is_notification_category_enabled(notification):
                if self.USE_MAIN_LOOP:
                    await self._send_notification(notification)
                else:
                    await self.loop.run_in_executor(self.executors, self._send_notification_from_executor, notification)
        except Exception as e:
            self.logger.exception(e, True, f"Exception when handling notification: {e}")

    async def _initialize_impl(self, backtesting_enabled, edited_config) -> bool:
        # make sure to always create the notification channel
        channel = await self._create_notification_channel_if_not_existing()
        if await abstract_service_user.AbstractServiceUser._initialize_impl(self, backtesting_enabled, edited_config):
            self.services = [service.instance() for service in self.REQUIRED_SERVICES]
            await self._subscribe_to_notification_channel(channel)
            return True
        return False

    async def _subscribe_to_notification_channel(self, channel):
        await channel.new_consumer(self._notification_callback)
        self.logger.debug("Registered as notification consumer")

    async def _order_notification_callback(self, exchange, exchange_id, cryptocurrency, symbol, order,
                                           update_type, is_from_bot):
        exchange_manager = trading_api.get_exchange_manager_from_exchange_id(exchange_id)
        # Do not notify on existing pre-start orders
        if is_from_bot and not trading_api.get_is_backtesting(exchange_manager):
            order_identifier = f"{exchange}_{order[trading_enums.ExchangeConstantsOrderColumns.ID.value]}"
            # find the last notification for this order if any
            linked_notification = self.previous_notifications_by_identifier[order_identifier] \
                if order_identifier in self.previous_notifications_by_identifier else None
            await self._handle_order_notification(order, linked_notification, order_identifier,
                                                  exchange_manager, exchange)

    async def _handle_order_notification(self, dict_order, linked_notification, order_identifier,
                                         exchange_manager, exchange):
        notification = None
        order_status = trading_api.parse_order_status(dict_order)

        if order_status is trading_enums.OrderStatus.OPEN:
            notification = notifications.OrderCreationNotification(linked_notification, dict_order, exchange)
            # update last notification for this order
            self.previous_notifications_by_identifier[order_identifier] = notification
        else:
            is_simulated = trading_api.is_trader_simulated(exchange_manager)
            if order_status is trading_enums.OrderStatus.EXPIRED or \
                order_status is trading_enums.OrderStatus.CANCELED or \
                    (order_status is trading_enums.OrderStatus.CLOSED
                     and dict_order[trading_enums.ExchangeConstantsOrderColumns.FILLED.value] == 0):
                notification = notifications.OrderEndNotification(linked_notification, None, exchange, [dict_order],
                                                                  None, None, None, False, is_simulated)
            elif order_status in (trading_enums.OrderStatus.CLOSED, trading_enums.OrderStatus.FILLED):
                _, profitability_percent, profitability_diff, _, _ = \
                    trading_api.get_profitability_stats(exchange_manager)
                order_profitability = trading_api.get_order_profitability(
                    exchange_manager, dict_order[trading_enums.ExchangeConstantsOrderColumns.ID.value])
                notification = notifications.OrderEndNotification(linked_notification, dict_order,
                                                                  exchange, [], order_profitability,
                                                                  profitability_percent, profitability_diff, True,
                                                                  is_simulated)
            # remove order from previous_notifications_by_identifier: no more notification from it to be received
            if order_identifier in self.previous_notifications_by_identifier:
                self.previous_notifications_by_identifier.pop(order_identifier)
        await self._notification_callback(notification)

    async def _subscribe_to_order_channel(self, exchange_id):
        try:
            await trading_api.subscribe_to_order_channel(self._order_notification_callback, exchange_id)
        except KeyError:
            self.logger.error("No order channel to subscribe to: impossible to send order notifications")

    @classmethod
    def is_enabled(cls, config):
        if cls.NOTIFICATION_TYPE_KEY is None:
            cls.get_logger().warning(f"{cls.get_name()}.NOTIFICATION_TYPE_KEY is not set, it has to be set to identify "
                                     f"and activate this notifier.")
        return cls.NOTIFICATION_TYPE_KEY in AbstractNotifier._get_activated_notification_keys(config)

    @staticmethod
    def _get_activated_notification_keys(config):
        if constants.CONFIG_CATEGORY_NOTIFICATION in config \
                and constants.CONFIG_NOTIFICATION_TYPE in config[constants.CONFIG_CATEGORY_NOTIFICATION]:
            return config[constants.CONFIG_CATEGORY_NOTIFICATION][constants.CONFIG_NOTIFICATION_TYPE]
        return []

    @staticmethod
    async def _create_notification_channel_if_not_existing() -> channels.Channel:
        try:
            return channels.get_chan(service_channels.NotificationChannel.get_name())
        except KeyError:
            channel = await channel_creator.create_channel_instance(service_channels.NotificationChannel,
                                                                    channels.set_chan)
            await channel.register_producer(service_channels.NotificationChannelProducer.instance(channel))
            return channel

    def _is_notification_category_enabled(self, notification):
        return constants.CONFIG_CATEGORY_NOTIFICATION in self.config and \
               self.config[constants.CONFIG_CATEGORY_NOTIFICATION].get(notification.category.value, True)



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.api as trading_api

import octobot_services.interfaces as interfaces


def get_global_profitability():
    simulated_global_profitability = 0
    real_global_profitability = 0
    simulated_no_trade_profitability = 0
    real_no_trade_profitability = 0
    simulated_full_origin_value = 0
    real_full_origin_value = 0
    market_average_profitability = None
    has_real_trader = False
    has_simulated_trader = False

    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):

            current_value, _, _, market_average_profitability, initial_portfolio_current_profitability = \
                trading_api.get_profitability_stats(exchange_manager)

            if trading_api.is_trader_simulated(exchange_manager):
                simulated_full_origin_value += trading_api.get_origin_portfolio_value(exchange_manager)
                simulated_global_profitability += current_value
                simulated_no_trade_profitability += initial_portfolio_current_profitability
                has_simulated_trader = True
            else:
                real_full_origin_value += trading_api.get_origin_portfolio_value(exchange_manager)
                real_global_profitability += current_value
                real_no_trade_profitability += initial_portfolio_current_profitability
                has_real_trader = True

    simulated_percent_profitability = simulated_global_profitability * 100 / simulated_full_origin_value \
        if simulated_full_origin_value > 0 else 0
    real_percent_profitability = real_global_profitability * 100 / real_full_origin_value \
        if real_full_origin_value > 0 else 0

    return has_real_trader, has_simulated_trader, \
        real_global_profitability, simulated_global_profitability, \
        real_percent_profitability, simulated_percent_profitability, \
        real_no_trade_profitability, simulated_no_trade_profitability, \
        market_average_profitability


def get_reference_market() -> str:
    # The reference market is the currency unit of the calculated quantity value
    return trading_api.get_reference_market(interfaces.get_global_config())



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.tentacles_management as tentacles_management

import octobot_services.interfaces as interfaces


class InterfaceFactory:
    def __init__(self, config):
        self.config = config

    @staticmethod
    def get_available_interfaces() -> list:
        return [interface_class
                for interface_class in tentacles_management.get_all_classes_from_parent(interfaces.AbstractInterface)
                if not tentacles_management.is_abstract_using_inspection_and_class_naming(interface_class)]

    async def create_interface(self, interface_class) -> interfaces.AbstractInterface:
        return interface_class(self.config)



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_services.managers as managers
import octobot_services.service_feeds as service_feeds


def create_service_feed_factory(config, main_async_loop, bot_id) -> service_feeds.ServiceFeedFactory:
    return service_feeds.ServiceFeedFactory(config, main_async_loop, bot_id)


def get_service_feed(service_feed_class, bot_id) -> service_feeds.AbstractServiceFeed:
    try:
        return service_feeds.ServiceFeeds.instance().get_service_feed(bot_id, service_feed_class.get_name())
    except TypeError:
        raise RuntimeError(f"can't get {service_feed_class} instance: service feed has not been properly created yet")


async def start_service_feed(service_feed: service_feeds.AbstractServiceFeed,
                             backtesting_enabled: bool,
                             edited_config: dict) -> bool:
    return await managers.ServiceFeedManager.start_service_feed(service_feed, backtesting_enabled, edited_config)


async def stop_service_feed(service_feed: service_feeds.AbstractServiceFeed) -> None:
    await managers.ServiceFeedManager.stop_service_feed(service_feed)


async def clear_bot_id_feeds(bot_id: str) -> None:
    service_feeds.ServiceFeeds.instance().clear_bot_id_feeds(bot_id)



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums as common_enums

import octobot_services.enums as enums


class Notification:
    def __init__(self, text: str, title: str, markdown_text: str, sound: enums.NotificationSound,
                 markdown_format: common_enums.MarkdownFormat,
                 level: enums.NotificationLevel, category: enums.NotificationCategory,
                 linked_notification):
        self.text = text
        self.markdown_text = markdown_text
        self.title = title
        self.level = level
        self.markdown_format = markdown_format
        self.linked_notification = linked_notification
        self.category = category
        self.sound = sound

        # Used to identify previous notification related elements when necessary ex: a tweet to reply to
        self.metadata = {}

    def __repr__(self):
        return f"[Notification] title: {self.title}, text: {self.text}, level: {self.level}, " \
               f"markdown_format: {self.markdown_format.name}, category: {self.category.value}, " \
               f"linked_notification: {self.linked_notification}"



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import abc


class InitializableWithPostAction:
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.is_initialized = False

    # calls initialize_impl if not initialized
    async def initialize(self, *args) -> bool:
        if not self.is_initialized:
            if await self._initialize_impl(*args):
                await self._post_initialize(args)
                self.is_initialized = True
                return True
            return False
        return False

    @abc.abstractmethod
    async def _initialize_impl(self, *args) -> bool:
        raise NotImplementedError("initialize_impl not implemented")

    # Implement _post_initialize if anything specific has to be done after initialize and before start
    async def _post_initialize(self, *args) -> bool:
        return True




#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.channels_name as channels_names
import async_channel.channels as channels
import octobot_services.interfaces as interfaces
import octobot_services.managers as managers


def initialize_global_project_data(bot_api: object, project_name: str, project_version: str) -> None:
    interfaces.AbstractInterface.initialize_global_project_data(bot_api, project_name, project_version)


def create_interface_factory(config: dict) -> interfaces.InterfaceFactory:
    return interfaces.InterfaceFactory(config)


def is_enabled(interface_class: interfaces.AbstractInterface) -> bool:
    return interface_class.enabled


async def send_user_command(bot_id, subject, action, data, wait_for_processing=False) -> bool:
    try:
        channel = channels.get_chan(channels_names.OctoBotUserChannelsName.USER_COMMANDS_CHANNEL.value)
        await channel.get_internal_producer().send(
            bot_id=bot_id,
            subject=subject,
            action=action,
            data=data
        )
        if wait_for_processing:
            producers = channel.producers
            if channel.internal_producer is not None:
                producers.append(channel.internal_producer)
            await asyncio.gather(*(producer.wait_for_processing() for producer in producers))
        return True
    except KeyError:
        return False


def is_enabled_in_backtesting(interface_class) -> bool:
    return all(service.BACKTESTING_ENABLED for service in interface_class.REQUIRED_SERVICES)


def is_interface_relevant(config, interface_class, backtesting_enabled):
    return is_enabled(interface_class) and \
           all(service.get_is_enabled(config) for service in interface_class.REQUIRED_SERVICES) and \
           (not backtesting_enabled or (backtesting_enabled and is_enabled_in_backtesting(interface_class)))


def disable_interfaces(interface_identifier: str) -> int:
    disabled_interfaces = 0
    normalized_identifier = interface_identifier.lower()
    for interface_class in interfaces.InterfaceFactory.get_available_interfaces():
        if normalized_identifier in interface_class.__name__.lower():
            interface_class.enabled = False
            disabled_interfaces += 1
    return disabled_interfaces


# Return the list of started interfaces
async def start_interfaces(interfaces: list) -> list:
    return await managers.start_interfaces(interfaces)


async def stop_interfaces(interfaces: list) -> None:
    await managers.stop_interfaces(interfaces)



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.channels as channels
import async_channel.producer as producer
import async_channel.consumer as consumer


class UserCommandsChannelConsumer(consumer.SupervisedConsumer):
    pass


class UserCommandsChannelProducer(producer.Producer):
    async def send(self, bot_id, subject, action, data=None):
        for consumer in self.channel.get_consumer_from_filters({"bot_id": bot_id, "subject": subject}):
            await consumer.queue.put({
                "bot_id": bot_id,
                "subject": subject,
                "action": action,
                "data": data
            })


class UserCommandsChannel(channels.Channel):
    PRODUCER_CLASS = UserCommandsChannelProducer
    CONSUMER_CLASS = UserCommandsChannelConsumer



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_services.services as services


async def stop_services():
    for service_instance in _get_service_instances():
        try:
            logging.get_logger(__name__).debug(f"Stopping {service_instance.get_name()} ...")
            await service_instance.stop()
            logging.get_logger(__name__).debug(f"Stopped {service_instance.get_name()}")
        except Exception as e:
            raise e


def _get_service_instances():
    return [service_class.instance() for service_class in services.ServiceFactory.get_available_services()]



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

# Config
CONFIG_CATEGORY_SERVICES = "services"
CONFIG_SERVICE_INSTANCE = "service_instance"

# Interfaces
CONFIG_INTERFACES = "interfaces"
CONFIG_INTERFACES_WEB = "web"
CONFIG_INTERFACES_TELEGRAM = "telegram"

# Service feeds
FEED_METADATA = "metadata"

# Telegram
CONFIG_TELEGRAM = "telegram"
CONFIG_TOKEN = "token"
CONFIG_TELEGRAM_CHANNEL = "telegram-channels"
MESSAGE_PARSE_MODE = 'parse_mode'
CONFIG_TELEGRAM_ALL_CHANNEL = "*"
CONFIG_GROUP_MESSAGE = "group-message"
CONFIG_GROUP_MESSAGE_DESCRIPTION = "group-message-description"
CONFIG_USERNAMES_WHITELIST = "usernames-whitelist"
CONFIG_CHAT_ID = "chat-id"

CONFIG_TELEGRAM_API = "telegram-api"
CONFIG_API = "telegram-api"
CONFIG_API_HASH = "telegram-api-hash"
CONFIG_TELEGRAM_PHONE = "telegram-phone"
CONFIG_TELEGRAM_PASSWORD = "telegram-password"
CONFIG_MESSAGE_CONTENT = "message-content"
CONFIG_MESSAGE_SENDER = "message-sender"
CONFIG_IS_GROUP_MESSAGE = "is-group-message"
CONFIG_IS_CHANNEL_MESSAGE = "is-channel-message"
CONFIG_IS_PRIVATE_MESSAGE = "is-private-message"
CONFIG_MEDIA_PATH = "media-path"

# Web
CONFIG_WEB = "web"
CONFIG_WEB_IP = "ip"
CONFIG_WEB_PORT = "port"
CONFIG_WEB_REQUIRES_PASSWORD = "require-password"
CONFIG_WEB_PASSWORD = "password"
CONFIG_AUTO_OPEN_IN_WEB_BROWSER = "auto-open-in-web-browser"
ENV_WEB_PORT = "WEB_PORT"
ENV_WEB_ADDRESS = "WEB_ADDRESS"
ENV_CORS_ALLOWED_ORIGINS = "CORS_ALLOWED_ORIGINS"
ENV_AUTO_OPEN_IN_WEB_BROWSER = "AUTO_OPEN_IN_WEB_BROWSER"
DEFAULT_SERVER_IP = '0.0.0.0'
DEFAULT_SERVER_PORT = 5001

# Webhook
CONFIG_WEBHOOK = "webhook"
CONFIG_ENABLE_NGROK = "enable-ngrok"
CONFIG_NGROK_TOKEN = "ngrok-token"
CONFIG_WEBHOOK_SERVER_IP = "webhook-bind-ip"
CONFIG_WEBHOOK_SERVER_PORT = "webhook-bind-port"
ENV_WEBHOOK_PORT = "WEBHOOK_PORT"
ENV_WEBHOOK_ADDRESS = "WEBHOOK_ADDRESS"
DEFAULT_WEBHOOK_SERVER_IP = '127.0.0.1'
DEFAULT_WEBHOOK_SERVER_PORT = 9000

# GPT
CONFIG_GPT = "GPT"
CONIG_OPENAI_SECRET_KEY = "openai-secret-key"
ENV_OPENAI_SECRET_KEY = "OPENAI_SECRET_KEY"
ENV_GPT_MODEL = "GPT_MODEL"
ENV_GPT_DAILY_TOKENS_LIMIT = "GPT_DAILY_TOKEN_LIMIT"

# Google
CONFIG_GOOGLE = "google"
CONFIG_TREND_TOPICS = "trends"
CONFIG_TREND = "trend"
CONFIG_TREND_DESCRIPTION = "trend_description"
CONFIG_TREND_HISTORY_TIME = "relevant_history_months"

# TradingView
CONFIG_TRADING_VIEW = "trading-view"
CONFIG_REQUIRE_TRADING_VIEW_TOKEN = "require-token"
CONFIG_TRADING_VIEW_TOKEN = "token"

# Twitter
CONFIG_TWITTERS_ACCOUNTS = "accounts"
CONFIG_TWITTERS_HASHTAGS = "hashtags"
CONFIG_TWITTER = "twitter"
CONFIG_TWITTER_API_INSTANCE = "twitter_api_instance"
CONFIG_TWEET = "tweet"
CONFIG_TWEET_DESCRIPTION = "tweet_description"
CONFIG_TW_API_KEY = "api-key"
CONFIG_TW_API_SECRET = "api-secret"
CONFIG_TW_ACCESS_TOKEN = "access-token"
CONFIG_TW_ACCESS_TOKEN_SECRET = "access-token-secret"

# Reddit
CONFIG_REDDIT = "reddit"
CONFIG_REDDIT_SUBREDDITS = "subreddits"
CONFIG_REDDIT_ENTRY = "entry"
CONFIG_REDDIT_ENTRY_WEIGHT = "entry_weight"
CONFIG_REDDIT_CLIENT_ID = "client-id"
CONFIG_REDDIT_CLIENT_SECRET = "client-secret"
CONFIG_REDDIT_PASSWORD = "password"
CONFIG_REDDIT_USERNAME = "username"

# Notifications
CONFIG_CATEGORY_NOTIFICATION = "notification"
CONFIG_NOTIFICATION_TYPE = "notification-type"

# Interfaces

PAID_FEES_STR = "Paid fees"

# external resources
EXTERNAL_RESOURCE_CURRENT_USER_FORM = "current-user-feedback-form"
EXTERNAL_RESOURCE_PUBLIC_ANNOUNCEMENTS = "public-announcements"



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.constants as constants

import octobot_services.interfaces as interfaces
import octobot_trading
import octobot_trading.api as trading_api


def _merge_portfolio_values(portfolio1, portfolio2):
    for key, value in portfolio2.items():
        if key in portfolio1:
            portfolio1[key] += value
        else:
            portfolio1[key] = value
    return portfolio1


def get_portfolio_holdings():
    real_currency_portfolio = {}
    simulated_currency_portfolio = {}

    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):

            trader_currencies_values = trading_api.get_current_holdings_values(exchange_manager)
            if trading_api.is_trader_simulated(exchange_manager):
                _merge_portfolio_values(simulated_currency_portfolio, trader_currencies_values)
            else:
                _merge_portfolio_values(real_currency_portfolio, trader_currencies_values)
    return real_currency_portfolio, simulated_currency_portfolio


def get_portfolio_current_value():
    simulated_value = 0
    real_value = 0
    has_real_trader = False
    has_simulated_trader = False

    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):

            current_value = trading_api.get_current_portfolio_value(exchange_manager)

            # current_value might be 0 if no trades have been made / canceled => use origin value
            if current_value == 0:
                current_value = trading_api.get_origin_portfolio_value(exchange_manager)

            if trading_api.is_trader_simulated(exchange_manager):
                simulated_value += current_value
                has_simulated_trader = True
            else:
                real_value += current_value
                has_real_trader = True

    return has_real_trader, has_simulated_trader, real_value, simulated_value


def _get_portfolios():
    simulated_portfolios = []
    real_portfolios = []

    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            if trading_api.is_trader_simulated(exchange_manager):
                simulated_portfolios.append(trading_api.get_portfolio(exchange_manager))
            else:
                real_portfolios.append(trading_api.get_portfolio(exchange_manager))
    return real_portfolios, simulated_portfolios


def _merge_portfolios(base_portfolio, to_merge_portfolio):
    for currency, asset in to_merge_portfolio.items():
        if currency not in base_portfolio:
            base_portfolio[currency] = {
                constants.PORTFOLIO_AVAILABLE: decimal.Decimal(0),
                constants.PORTFOLIO_TOTAL: decimal.Decimal(0)
            }

        base_portfolio[currency][constants.PORTFOLIO_AVAILABLE] += asset.available
        base_portfolio[currency][constants.PORTFOLIO_TOTAL] += asset.total


def get_global_portfolio_currencies_amounts():
    real_portfolios, simulated_portfolios = _get_portfolios()
    real_global_portfolio = {}
    simulated_global_portfolio = {}

    for portfolio in simulated_portfolios:
        _merge_portfolios(simulated_global_portfolio, portfolio)

    for portfolio in real_portfolios:
        _merge_portfolios(real_global_portfolio, portfolio)

    return real_global_portfolio, simulated_global_portfolio


def get_global_portfolio_currencies_values() -> dict:
    return trading_api.get_global_portfolio_currencies_values(
        interfaces.get_exchange_managers()
        )
                

def trigger_portfolios_refresh():
    at_least_one = False
    for exchange_manager in interfaces.get_exchange_managers():
        if trading_api.is_trader_existing_and_enabled(exchange_manager):
            at_least_one = True
            interfaces.run_in_bot_main_loop(trading_api.refresh_real_trader_portfolio(exchange_manager))

    if not at_least_one:
        raise RuntimeError("no real trader to update.")



#  Drakkar-Software OctoBot-Services
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc

import octobot_commons.configuration as configuration
import octobot_commons.singleton as singleton

import octobot_services.constants as constants


class AbstractService(singleton.Singleton):
    __metaclass__ = abc.ABCMeta

    BACKTESTING_ENABLED = False

    def __init__(self):
        super().__init__()
        self.logger = None
        self.config = None
        self.edited_config = None
        self._created = True
        self._healthy = False
        self._has_been_created = False

    def set_has_been_created(self, value):
        self._has_been_created = value

    def get_has_been_created(self):
        return self._has_been_created

    def is_healthy(self):
        return self._healthy

    @classmethod
    def get_name(cls):
        return cls.__name__

    def get_fields_description(self) -> dict:
        """
        :return: the service configuration keys with their description
        """
        return {}

    def get_default_value(self) -> dict:
        """
        :return: the service configuration keys with their default value
        """
        return {}

    def get_required_config(self) -> list:
        """
        :return: the list of required (not optional) configuration keys
        """
        return []

    def get_read_only_info(self) -> dict:
        return {}

    @classmethod
    def get_help_page(cls) -> str:
        """
        :return: the url of the help page with this service
        """
        return ""

    # Override this method if a user is to be registered in this service (ie: TelegramService)
    def register_user(self, user_key):
        pass

    # Override this method if a service feed is located in this service (ie: TelegramService)
    async def start_service_feed(self):
        pass

    # Override this method to know if an updater is already running
    def is_running(self):
        return False

    # Returns true if all the service has an instance in config
    @staticmethod
    @abc.abstractmethod
    def is_setup_correctly(config):
        raise NotImplementedError("is_setup_correctly not implemented")

    # Override this method to perform additional checks
    @staticmethod
    def get_is_enabled(config):
        return True

    # implement locally if the service has thread(s) to stop
    async def stop(self):
        pass

    # implement locally if the service shouldn't raise warning at startup if configuration is not set
    @staticmethod
    def get_should_warn():
        return True

    # Returns true if all the configuration is available
    @abc.abstractmethod
    def has_required_configuration(self):
        raise NotImplementedError("has_required_configuration not implemented")

    # Returns the service's endpoint
    @abc.abstractmethod
    def get_endpoint(self) -> None:
        raise NotImplementedError("get_endpoint not implemented")

    # Called to put in the right service in config
    @abc.abstractmethod
    def get_type(self) -> None:
        raise NotImplementedError("get_type not implemented")

    # Called after service setup
    @abc.abstractmethod
    async def prepare(self) -> None:
        raise NotImplementedError("prepare not implemented")

    # Called by say_hello after service is prepared, return relevant service information and a boolean for
    # success or failure
    @abc.abstractmethod
    def get_successful_startup_message(self):
        raise NotImplementedError("get_successful_startup_message not implemented")

    def get_website_url(self) -> str:
        # Override with the real url
        return ""

    def get_brand_name(self):
        return self.get_type()

    def get_logo(self):
        # Default services logos are fount using https://github.com/edent/SuperTinyIcons
        # Override to customize this behavior
        return f"https://raw.githubusercontent.com/edent/SuperTinyIcons/master/images/svg/{self.get_brand_name()}.svg"

    def check_required_config(self, config):
        return all(key in config for key in self.get_required_config()) and \
            not configuration.has_invalid_default_config_value(*(config[key] for key in self.get_required_config()))

    def log_connection_error_message(self, e):
        self.logger.error(f"{self.get_name()} is failing to connect, please check your internet connection: {e}")

    async def say_hello(self):
        message, self._healthy = self.get_successful_startup_message()
        if self._healthy and message:
            self.logger.info(message)
        return self._healthy

    def save_service_config(self, service_key, service_config, update=False):
        """
        Save the service's config into the user config file
        :param service_key: the key of the service config in file
        :param service_config: the updated config
        :param update: when true, the service configuration dict will be updated using the new data, it will
        be replaced otherwise
        :return: None
        """
        if update and service_key in self.edited_config.config[constants.CONFIG_CATEGORY_SERVICES]:
            self.edited_config.config[constants.CONFIG_CATEGORY_SERVICES][service_key].update(service_config)
        else:
            self.edited_config.config[constants.CONFIG_CATEGORY_SERVICES][service_key] = service_config
        self.edited_config.save()
